#!/usr/bin/env python
# -*- coding: gb2312 -*-
import ctypes
import datetime
import os
import random
import sys
import time
from datetime import date

sys.path.append(".")
import server, security, record, dataprovider

class TestSDK:
    def ConnectServer(self):
        self.myServer = server.Server('127.0.0.1', 5673, '', 5673)
        self.myServer.connect()
        self.myServer.login('admin', 'admin')
        self.myTagMgr = self.myServer.tag_mgr
        self.myDateProvider = self.myServer.data_provider

    def AddTag(self):
        attr = {'tagname': 'testtag', 'tagtype': 'int8'}
        try:
            self.myTagMgr.add_tag(**attr)
        except Exception as e:
            print("add_tag error: ", e.errcode)

    def AddRecord(self):
        try:
            self.myTag = self.myTagMgr.get_tag('testtag')
        except Exception as e:
            print(e)
        hdrecord = record.Record(int(time.strftime('%m%d%H%M%S', time.localtime())), 0)
        hdrecord.tagtype = self.myTag.tagtype
        hdrecord.quality = 192
        hdrecord.value = 123
        try:
            self.myTag.save_snapshot(hdrecord)
        except Exception as e:
            print("save_snapshot error: ", e.errcode)

    def AddTags(self):
        attr = [{'tagname': 'tag_int16_0', 'tagtype': 'int8'},
                {'tagname': 'tag_int16_1', 'tagtype': 'int8'},
                {'tagname': 'tag_int16_2', 'tagtype': 'int8'},
                {'tagname': 'tag_int16_3', 'tagtype': 'int8'},
                {'tagname': 'tag_int16_4', 'tagtype': 'int8'},
                {'tagname': 'tag_int16_5', 'tagtype': 'int8'},
                {'tagname': 'tag_int16_6', 'tagtype': 'int8'},
                {'tagname': 'tag_int16_7', 'tagtype': 'int8'},
                {'tagname': 'tag_int16_8', 'tagtype': 'int8'},
                {'tagname': 'tag_int16_9', 'tagtype': 'int8'}]
        try:
            self.myTagMgr.add_tags(attr)
        except Exception as e:
            print("add_tag error: ", e.errcode)

    def AddRecords(self):
        tagnames = []
        hdrecord_list = []
        snapshot_list = []
        for i in range(10):
            tagnames.append('tag_int16_' + str(i))
        for tagname in tagnames:
            try:
                self.myTag = self.myTagMgr.get_tag(tagname)
            except Exception as e:
                print(e)
            hdrecord = record.Record(int(time.strftime('%m%d%H%M%S', time.localtime())), 0)
            hdrecord.tagtype = self.myTag.tagtype
            hdrecord.quality = 192
            hdrecord.value = random.randint(0, 10)
            hdrecord_list.append(hdrecord)
        try:
            ret = self.myTagMgr.get_tagIDs(tagnames)
        except Exception as e:
            print("get_tagIDs error: ", e.errcode)
        for i in range(10):
            snapshot_list.append((ret[1][i], hdrecord_list[i]))
        try:
            self.myTag.save_snapshots(snapshot_list)
        except Exception as e:
            print("save_snapshot error: ", e.errcode)

    def QuerySnapShot(self):
        try:
            hdrecord = self.myTag.get_snapshot()
        except Exception as e:
            print("get_snapshot error: ", e.errcode)
        else:
            print("snapshot quality: ", hdrecord.quality, "snapshot value: ", hdrecord.value, ", timestamp: ", hdrecord.sec)

    def QuerySnapShot2(self):
        tagnames = []
        tagtypes = []
        for i in range(10):
            tagnames.append('tag_int16_' + str(i))
        try:
            ret = self.myTagMgr.get_tagIDs(tagnames)
        except Exception as e:
            print("get_tagIDs error: ", e.errcode)
        tagIDs = ret[1]
        len_tagnames = len(tagnames)
        for i in range(len_tagnames):
            tagtypes.append(self.myTagMgr.get_tagType(tagIDs[i]))

        try:
            ret1 = self.myTag.get_snapshots(tagIDs, tagtypes)
        except Exception as e:
            print("get_tagIDs error: ", e.errcode)
        reclist = ret1[1]
        len_reclist = len(reclist)
        for i in range(len_reclist):
            print("snapshot quality: ", reclist[i].quality, "snapshot value: ", reclist[i].value,
                  ", timestamp: ", reclist[i].sec, "snapshot type: ", reclist[i].tagtype)

    def QuerySnapShots(self):
        tagnames = []
        for i in range(10):
            tagnames.append('tag_int16_' + str(i))

        try:
            ret = self.myTagMgr.get_tagIDs(tagnames)
        except Exception as e:
            print("get_tagIDs error: ", e.errcode)

        try:
            res = self.myDateProvider.get_snapshot_by_IDs(ret[1])
        except Exception as e:
            print("get_snapshot_by_IDs error: ", e.errcode)
        else:
            for i in range(10):
                print("tagname: ", tagnames[i], "snapshot value: ", res[1][i].value)

    def GetInterRecordByMode(self):
        tagname = 'tag_int16_0'
        try:
            self.myTag = self.myTagMgr.get_tag(tagname)
        except Exception as e:
            print(e)
        self.dattime = datetime.datetime(2023, 10, 11, 16, 31, 15, 999)

        try:
            hdrecord = self.myTag.get_inter_record_by_mode(0, timestamp=self.dattime)
        except Exception as e:
            print("get_inter_record_by_mode error: ", e.errcode)
        else:
            print("snapshot quality: ", hdrecord.quality, "snapshot value: ", hdrecord.value, ", timestamp: ",
                  hdrecord.sec)

    def GetInterRecordsByMode(self):
        tagname = 'tag_int16_0'
        try:
            self.myTag = self.myTagMgr.get_tag(tagname)
        except Exception as e:
            print(e)
        times = []
        for second in range(40, 50):
            self.dattime = datetime.datetime(2023, 10, 12, 9, 24, second)
            times.append(self.dattime)
        try:
            ret = self.myTag.get_inter_records_by_mode(0, times)
            hdrecords = ret[1]
        except Exception as e:
            print("get_inter_records_by_mode error: ", e.errcode)
        for hdrecord in hdrecords:
            print("snapshot quality: ", hdrecord.quality, "snapshot value: ", hdrecord.value, ", timestamp: ",
                  hdrecord.sec)

    def GetRawRecordsNumber(self):
        tagname = 'tag_int16_0'
        try:
            self.myTag = self.myTagMgr.get_tag(tagname)
        except Exception as e:
            print(e)
        sttime = datetime.datetime(2023, 10, 12, 9,24, 40)
        edtime = datetime.datetime(2023, 10, 12, 9, 24, 50)
        num_hdrecored = self.myTag.get_raw_records_number(starttime=sttime, endtime=edtime)
        print("number of hdrecords in (", sttime, ", ", edtime, "]: ", num_hdrecored)
        pass

    def ModifyRecord(self):
        tagname = 'tag_int16_0'
        try:
            self.myTag = self.myTagMgr.get_tag(tagname)
        except Exception as e:
            print(e)
        self.hdrecord = record.Record(1697421143, 0)
        self.hdrecord.value = 22
        self.hdrecord.quality = 192
        self.hdrecord.tagtype = 'int8'
        self.myTag.modify_record(self.hdrecord)

    def Disconnet(self):
        self.myServer.disconnect()


if __name__ == '__main__':
    testA = TestSDK()
    print("PythonSDK_V3 sample..")
    testA.ConnectServer()
    time.sleep(3)
    testA.ModifyRecord()
    testA.Disconnet()