#!/usr/bin/env python
# -*- coding: gb2312 -*-

from ctypes import *
import hyperdb
import tagmgr
import usergroupmgr
import securitymgr
import alarmgroupmgr
import sys
import collectormgr
import dataprovider


## Each object derived from class Server should be corresponding to a
#  iHyperDB Server
class Server(object):
    ## Construction function of class Server
    # @param ip: ip address with string formation, such as "168.2.4.7"
    # @param port: port number which small than 65535 and is not used by other program 
    def __init__(self, address, port, bakaddress='', bakport=0):
        if (type(address) != str) or (type(port) != int):
            raise TypeError
        if (type(bakaddress) != str) or (type(bakport) != int):
            raise TypeError
        self.address = address
        self.port = port
        self.bakaddress = bakaddress
        self.bakport = bakport
        hyperdb.load_library()
        self.tag_mgr = tagmgr.TagMgr(self)
        self.usergroup_mgr = usergroupmgr.UserGroupMgr()
        self.security_mgr = securitymgr.SecurityMgr()
        self.alarmgroup_mgr = alarmgroupmgr.AlarmGroupMgr()
        self.collector_mgr = collectormgr.CollectorMgr()
        self.data_provider = dataprovider.DataProvider(self)

    ## Connect to a server
    # @param timeout: maximal time that connect operator is allowed
    def connect(self, timeout=3):
        if (type(timeout) != int):
            raise TypeError
        conn_config = hyperdb.HD3Connection()
        conn_config.szAddress = self.address.encode('utf-8')
        conn_config.nPort = c_uint16(self.port)
        conn_config.nTimeout = c_int32(timeout)

        if self.bakaddress != '':
            conn_config.szBakAddress = self.bakaddress.encode('utf-8')
        else:
            conn_config.szBakAddress = ''.encode('utf-8')

        if (self.bakport != 0):
            conn_config.nBakPort = c_uint16(self.bakport)
        else:
            conn_config.nBakPort = 0

        hyperdb.api.nt3_connect.argtypes = [POINTER(hyperdb.HD3Connection)]
        retcode = hyperdb.api.nt3_connect(byref(conn_config))

        if (retcode != hyperdb.hd_sucess):
            print(retcode)
            raise hyperdb.HDError(retcode)

    ## disconnect from a server        
    def disconnect(self):

        retcode = hyperdb.api.nt3_disconnect()

        if (hyperdb.hd_sucess != retcode):
            raise hyperdb.HDError(retcode)

    ## Set timeout for sending commands
    # @param timeout: timeout for sending commands except for connection
    # (set_timeout -> set_request_timeout)
    def set_request_timeout(self, timeout=3):
        if (type(timeout) != int):
            raise TypeError

        timeout = c_int32(timeout)

        try:
            hyperdb.api.ut3_set_request_timeout.argtypes = [c_int32]
            hyperdb.api.ut3_set_request_timeout(timeout)
        except Exception as e:
            print(e.errcode)
            raise hyperdb.HDError(e.errcode)

    ## Login a server
    # @param username: user name
    # @param password: user's password which should be longer than 5 characters
    def login(self, username, password):
        if (type(username) != str) or (type(password) != str):
            raise TypeError

        name_buf = username
        password_buf = password

        hyperdb.api.sc3_login.argtypes = [c_char_p, c_char_p]
        retcode = hyperdb.api.sc3_login(name_buf.encode('utf-8'), password_buf.encode('utf-8'))
        if (hyperdb.hd_sucess != retcode):
            raise hyperdb.HDError(retcode)

    # ַ
    ## set the charset
    # @brief: APIַʾAPIܵı롢APIı룩
    # @param charset: ַö
    # typedef enum _HD3_CHARSET{
    # HD3_CHARSET_SYS,  // ϵͳĬַ
    # HD3_CHARSET_ANSI,  // ANSIַGB2312,GBK,GB1830
    # HD3_CHARSET_UNICODE  // Unicodeַ
    # }HD3_CHARSET;
    # @raise exception: TypeError, hyperdb.HDError(e.errcode)
    def set_charset(self, charset):
        if (type(charset) is not int):
            raise TypeError

        charset = c_int32(charset)
        try:
            hyperdb.api.ut3_set_charset.argtypes = [c_int32]
            hyperdb.api.ut3_set_charset(charset)
        except Exception as e:
            print(e.errcode)
            raise hyperdb.HDError(e.errcode)

    ## get the charset
    # @brief: ȡAPIַʾAPIܵı롢APIı룩
    # @param charset: ַö
    # typedef enum _HD3_CHARSET{
    # HD3_CHARSET_SYS,  // ϵͳĬַ
    # HD3_CHARSET_ANSI,  // ANSIַGB2312,GBK,GB1830
    # HD3_CHARSET_UNICODE  // Unicodeַ
    # }HD3_CHARSET;
    # @raise exception: hyperdb.HDError(e.errcode)
    def get_charset(self):
        charset = c_int32()
        try:
            charset = hyperdb.api.ut3_get_charset()
        except Exception as e:
            print(e.errcode)
            raise hyperdb.HDError(EOFError.errcode)
        return charset

    ## set connection checking cycle
    # @brief üʱ䣬ڶڼAPĮ,λΪ
    # @param nInterval []
    # @version 03/24/2015 hd3 team initial version
    # @raise exception: hyperdb.HDError(e.errcode)
    def set_connection_checking_cycle(self, nInterval):
        if (type(nInterval) != int):
            raise TypeError

        nInterval = c_int32(nInterval)

        try:
            hyperdb.api.ut3_set_connection_checking_cycle.argtypes = [c_int32]
            hyperdb.api.ut3_set_connection_checking_cycle(nInterval)
        except Exception as e:
            print(e.errcode)
            raise hyperdb.HDError(e.errcode)

    ## get server ms time
    # ȡĻʱ䣬λΪ
    # @return: pMsTime
    # @raise exception: hyperdb.HDError(retcode)
    def get_server_ms_time(self):
        pMsTime = c_int64()
        hyperdb.api.ut3_get_server_ms_time.argtypes = [POINTER(c_int64)]
        retcode = hyperdb.api.ut3_get_server_ms_time(byref(pMsTime))
        if (hyperdb.hd_sucess != retcode):
            raise hyperdb.HDError(retcode)
        return pMsTime
